#include "obj.h"
#include "aliases.h"
#include "img.h"
#include "mem_arena.h"
#include "typed_list.h"
#include "utils.h"
#include <stdio.h>

#define TRIANGLE_VERTICES 3

typedef struct triangle_bbox TriangleBBox;
struct triangle_bbox {
  u64 x0;
  u64 y0;
  u64 x1;
  u64 y1;
};

typedef struct i64x2 V2i;
struct i64x2 {
  i64 x;
  i64 y;
};

typedef struct f32x3 V3f;
struct f32x3 {
  f32 x;
  f32 y;
  f32 z;
};

typedef struct u64x3 V3u;
struct u64x3 {
  u64 x;
  u64 y;
  u64 z;
};

typedef struct f32_3x3 M3x3f;
struct f32_3x3 {
  V3f row0;
  V3f row1;
  V3f row2;
};

internal void render_triangle(const Triangle *triangle, const Model *model,
                              Image *img, Colour colour, RenderType type);
internal void fill_triangle(Image *img, Vertex vertices[TRIANGLE_VERTICES],
                            Colour colour);
internal TriangleBBox get_triangle_bbox(const Image *img,
                                        Vertex vertices[TRIANGLE_VERTICES]);
internal V3f get_barycentric_coords(f32 d00, f32 d01, f32 d11, f32 denom,
                                    const V2i *ab, const V2i *ac,
                                    const V2i *ap);
internal void get_image_coordinates(const Vertex *vertex, const Image *img,
                                    u64 *x, u64 *y);
internal u64 ndc_to_image_coordinate(f32 value, u64 max);

Model load_obj_file(Arena *arena, const char *filename) {
  if (!arena) {
    return NULL_MODEL;
  }

  FILE *fp = fopen(filename, "r");
  if (!fp) {
    return NULL_MODEL;
  }

  Model model = (Model){
      .vertices = list_create(Vertex, arena),
      .triangles = list_create(Triangle, arena),
  };
  if (!(model.vertices) || !(model.triangles)) {
    return NULL_MODEL;
  }

  char line[8192];
  char identifier[8];
  Vertex vertex;
  Triangle triangle;
  f32 vx, vy, vz;
  u64 fp0, fp1, fp2;
  u64 ign_0_1, ign_0_2;
  u64 ign_1_1, ign_1_2;
  u64 ign_2_1, ign_2_2;
  while (fgets(line, 8191, fp) != NULL) {
    sscanf(line, "%s", identifier);
    if (strncmp(identifier, "v", 8) == 0) {
      sscanf(line + 2, "%f %f %f", &vx, &vy, &vz);
      vertex.x = vx;
      vertex.y = vy;
      vertex.z = vz;
      list_append(Vertex, arena, model.vertices, vertex);
    } else if (strncmp(identifier, "f", 8) == 0) {
      sscanf(line + 2, "%lu/%lu/%lu %lu/%lu/%lu %lu/%lu/%lu", &fp0, &ign_0_1,
             &ign_0_2, &fp1, &ign_1_1, &ign_1_2, &fp2, &ign_2_1, &ign_2_2);
      // OBJ indices start from 1
      triangle.p0 = fp0 - 1;
      triangle.p1 = fp1 - 1;
      triangle.p2 = fp2 - 1;
      list_append(Triangle, arena, model.triangles, triangle);
    }
  }

  return model;
}

void render_model(const Model *model, Image *img, Colour colour,
                  RenderType type) {
  Triangle triangle;

  for (u64 i = 0; i < model->triangles->count; ++i) {
    triangle = list_get(model->triangles, i);
    render_triangle(&triangle, model, img, colour, type);
  }
}

internal void render_triangle(const Triangle *triangle, const Model *model,
                              Image *img, Colour colour, RenderType type) {
  Vertex vertices[TRIANGLE_VERTICES] = {
      list_get(model->vertices, triangle->p0),
      list_get(model->vertices, triangle->p1),
      list_get(model->vertices, triangle->p2),
  };

  if (type == RENDER_TYPE_WIREFRAME) {
    Vertex v0, v1;
    u64 x0, y0, x1, y1;
    for (u64 i = 0; i < TRIANGLE_VERTICES; ++i) {
      v0 = vertices[i];
      v1 = vertices[(i + 1) % TRIANGLE_VERTICES];

      get_image_coordinates(&v0, img, &x0, &y0);
      get_image_coordinates(&v1, img, &x1, &y1);

      draw_line(img, x0, y0, x1, y1, colour);
    }
  } else if (type == RENDER_TYPE_FILLED) {
    fill_triangle(img, vertices, colour);
  }
}

internal void fill_triangle(Image *img, Vertex vertices[TRIANGLE_VERTICES],
                            Colour colour) {
  TriangleBBox bbox = get_triangle_bbox(img, vertices);

  u64 v0x, v0y, v1x, v1y, v2x, v2y;
  get_image_coordinates(&(vertices[0]), img, &v0x, &v0y);
  get_image_coordinates(&(vertices[1]), img, &v1x, &v1y);
  get_image_coordinates(&(vertices[2]), img, &v2x, &v2y);

  V2i ab = {.x = (i64)v1x - (i64)v0x, .y = (i64)v1y - (i64)v0y};
  V2i ac = {.x = (i64)v2x - (i64)v0x, .y = (i64)v2y - (i64)v0y};
  f32 d00 = (f32)ab.x * (f32)ab.x + (f32)ab.y + (f32)ab.y;
  f32 d01 = (f32)ab.x * (f32)ac.x + (f32)ab.y + (f32)ac.y;
  f32 d11 = (f32)ac.x * (f32)ac.x + (f32)ac.y + (f32)ac.y;
  f32 denom = (f32)d00 * (f32)d11 - (f32)d01 * (f32)d01;
  V2i ap;
  V3f coords;

  for (u64 y = bbox.y0; y < bbox.y1; ++y) {
    for (u64 x = bbox.x0; x < bbox.x1; ++x) {
      ap = (V2i){.x = (i64)x - (i64)v0x, .y = (i64)y - (i64)v0y};
      coords = get_barycentric_coords(d00, d01, d11, denom, &ab, &ac, &ap);
      if (coords.x < 0.0f || coords.y < 0.0f || coords.z < 0.0f) {
        continue;
      }
      set_pixel(img, x, y, colour);
    }
  }
}

internal TriangleBBox get_triangle_bbox(const Image *img,
                                        Vertex vertices[TRIANGLE_VERTICES]) {
  f32 x0 = min(vertices[0].x, min(vertices[1].x, vertices[2].x));
  f32 x1 = max(vertices[0].x, max(vertices[1].x, vertices[2].x));
  // NOTE (Abdelrahman): Because y is flipped, we use max for the minimum and
  // min for the maximum
  f32 y0 = max(vertices[0].y, max(vertices[1].y, vertices[2].y));
  f32 y1 = min(vertices[0].y, min(vertices[1].y, vertices[2].y));

  return (TriangleBBox){
      .x0 = ndc_to_image_coordinate(x0, img->width),
      .y0 = ndc_to_image_coordinate(0.0f - y0, img->height),
      .x1 = ndc_to_image_coordinate(x1, img->width),
      .y1 = ndc_to_image_coordinate(0.0f - y1, img->height),
  };
}

internal V3f get_barycentric_coords(f32 d00, f32 d01, f32 d11, f32 denom,
                                    const V2i *ab, const V2i *ac,
                                    const V2i *ap) {
  f32 d20 = (f32)ap->x * (f32)ab->x + (f32)ap->y + (f32)ab->y;
  f32 d21 = (f32)ap->x * (f32)ac->x + (f32)ap->y + (f32)ac->y;

  f32 v = (d11 * d20 - d01 * d21) / denom;
  f32 w = (d00 * d21 - d01 * d20) / denom;
  f32 u = 1.0f - v - w;

  return (V3f){v, w, u};
}

internal void get_image_coordinates(const Vertex *vertex, const Image *img,
                                    u64 *x, u64 *y) {
  *x = ndc_to_image_coordinate(vertex->x, img->width);
  *y = ndc_to_image_coordinate(0.0f - vertex->y, img->height);
}

internal u64 ndc_to_image_coordinate(f32 value, u64 max) {
  f32 result = (value + 1.0f) * max * 0.5f;
  return clamp((u64)result, 0, max);
}
